from typing import List, Tuple
import spacy
import sent_pattern
import pytest


nlp = spacy.load("en_core_web_lg")
nlp.add_pipe("span_noun")
nlp.add_pipe("sent_pattern")


test_data = {
    "texts":[
        "Apple’s new iPhone 14 lineup isn’t proving to be as popular as the company would have hoped for",
        "Apple also began to produce the iPhone 14 in India this week in order to shift some burden of manufacturing from factories in China",
        "A recent report published by JP Morgan suggested that Apple is looking to shift 25% of its iPhone production to India by 2025. ",
        "Tech author James Ball told the BBC it was `not a surprise` that Meta has had to agree to a serious pay-out but that it was `not that much` money to the tech giant.",
    ],
    "subject": [
        "Apple’s new iPhone 14 lineup",
        "Apple",
        "A recent report",
    ],
    "verb": [
        "proving",
        "began",
        "suggested",
        "told"
    ],
    "adjective":[
        "be",
        "produce",
        "looking"
    ],
    "pattern":[
        "SVC",
        "SV",
        "SVO",
        "SVOO",
    ]
}


def make_test_data(texts: List[str], answers: List[str])-> List[Tuple[str,str]]:
    data = []
    for text, ans in zip(texts, answers):
        data.append((text, ans))
    return data

subject_test_data = make_test_data(test_data["texts"], test_data["subject"])
@pytest.mark.parametrize(('text', 'subject'), subject_test_data)
def test_subject(text: str, subject: str):
    doc = nlp(text)
    pattern = doc._.sentpattern
    assert pattern.subject.root.text == subject


verb_test_data = make_test_data(test_data["texts"], test_data["verb"])
@pytest.mark.parametrize(('text', 'verb'), verb_test_data)
def test_verb(text:str, verb:str):
    doc = nlp(text)
    pattern = doc._.sentpattern
    assert pattern.verb.root.text == verb


pattern_test_data = make_test_data(test_data["texts"], test_data["adjective"])
@pytest.mark.parametrize(('text', 'adjective'), pattern_test_data)
def test_pattern(text:str, adjective:str):
    doc = nlp(text)
    pattern = doc._.sentpattern
    assert pattern.adjective.root.text == adjective



objects_data = {
    "texts": [
        "In this quickstart we’ll create a development shell with specific tools installed. ",
        ],
    "object":[
        "a development shell",
        ],
}
def make_test_object_data(texts: List[str], answers: List[str])-> List[Tuple[str,str]]:
    data = []
    for text, ans in zip(texts, answers):
        data.append((text, ans))
    return data
object_test_data = make_test_object_data(objects_data["texts"], objects_data["object"])
@pytest.mark.parametrize(('text', 'object'), object_test_data)
def test_object(text: str, object: str):
    doc = nlp(text)
    pattern = doc._.sentpattern
    if len(pattern.object.root) == 1:
        assert pattern.object.root[0].text == object
    else:
        assert pattern.object.root[0].text + " " + pattern.object.root[1].text == object



from sent_pattern.core.elements import Subject, Verb, RootObject, Adjective
from sent_pattern.core.type import FifthSpanDictType, FifthSpanDictTypeStr, FirstSpanDictType, FirstSpanDictTypeStr, FourthSpanDictType, FourthSpanDictTypeStr, SecondSpanDictType, SecondSpanDictTypeStr, ThirdSpanDictType, ThirdSpanDictTypeStr
from ..interface.Ipattern import IBaseSentencePattern
from typing import  Union
from spacy.symbols import ADJ,nsubj
from typing import TypeAlias
from sent_pattern.core.elements import Adjective,RootObject, RootElements



class FirstSentencePattern(IBaseSentencePattern):
    def __init__(self, subject: Subject, verb: Verb):
        self._subject = subject
        self._verb = verb
        self._span_dict:FirstSpanDictType = dict()
        self._span_dict_str:FirstSpanDictTypeStr =dict()

    @property
    def subject(self) -> "Subject": return self._subject
    @property
    def verb(self) -> "Verb": return self._verb
    
    @property
    def spans(self) -> FirstSpanDictType:
        if self._is_element_in_span_dict(self._span_dict):
            return self._span_dict
        self._span_dict["S"] = self._subject.span(self._subject.root)
        self._span_dict["V"] = self._verb.root
        return self._span_dict
    
    @property
    def spans_str(self) -> FirstSpanDictTypeStr:
        if self._is_element_in_span_dict(self._span_dict_str):
            return self._span_dict_str
        subject_span = self._subject.span(self._subject.root)
        self._span_dict_str["S"] = self._subject.span_str(subject_span)
        self._span_dict_str["V"] = self._verb.root.text
        return self._span_dict_str
    
    def _is_element_in_span_dict(self, dic: Union[FirstSpanDictType,FirstSpanDictTypeStr]) -> bool:
        if len(dic) == 0:
            return False
        return True
    
    @property
    def abbreviation(self) -> str:
        return "SV"


class SecondSentencePattern(IBaseSentencePattern):
    def __init__(self, subject: Subject, verb: Verb, adjective: Adjective):
        self._subject = subject
        self._verb = verb
        self._adjective = adjective
        self._adjective_root = adjective.root
        self._subject_root = subject.root
        self._verb_root = verb.root
        self._span_dict:SecondSpanDictType = dict()
        self._span_dict_str:SecondSpanDictTypeStr = dict()

    @property
    def subject(self) -> "Subject": return self._subject
    @property
    def verb(self) -> "Verb": return self._verb
    @property
    def adjective(self) -> "Adjective": return self._adjective
    
    @property
    def spans(self) -> SecondSpanDictType:
        if self._is_element_in_span_dict(self._span_dict):
            return self._span_dict
        self._span_dict["S"] = self._subject.span(self._subject_root)
        self._span_dict["V"] = self._verb_root
        self._span_dict["C"] = self._adjective.span(self._adjective_root)
        return self._span_dict
    
    @property
    def spans_str(self) -> SecondSpanDictTypeStr:
        if self._is_element_in_span_dict(self._span_dict_str):
            return self._span_dict_str
        subject_span = self._subject.span(self._subject.root)
        adjective_span = self._adjective.span(self._adjective_root)
        self._span_dict_str["S"] = self._subject.span_str(subject_span)
        self._span_dict_str["V"] = self._verb.root.text
        self._span_dict_str["C"] = self._adjective.span_str(adjective_span)
        return self._span_dict_str
    
    def _is_element_in_span_dict(self, dic:Union[SecondSpanDictType,SecondSpanDictTypeStr]) -> bool:
        if len(dic) == 0:
            return False
        return True
    
    @property
    def abbreviation(self) -> str:
        return "SVC"



class ThirdSentencePattern(IBaseSentencePattern):
    def __init__(self, subject: Subject, verb: Verb, object: RootObject):
        self._subject = subject
        self._verb = verb
        self._object = object
        self._object_root = object.root
        self._subject_root = subject.root
        self._verb_root = verb.root
        self._span_dict:ThirdSpanDictType = dict()
        self._span_dict_str:ThirdSpanDictTypeStr = dict()

    @property
    def subject(self) -> "Subject": return self._subject
    @property
    def verb(self) -> "Verb": return self._verb
    @property
    def object(self) -> "RootObject": return self._object

    @property
    def spans(self) -> ThirdSpanDictType:
        if self._is_element_in_span_dict(self._span_dict):
            return self._span_dict
        self._span_dict["S"] = self._subject.span(self._subject_root)
        self._span_dict["V"] = self._verb_root
        self._span_dict["O"] = self._object.span(self._object_root)
        return self._span_dict
    
    @property
    def spans_str(self) -> ThirdSpanDictTypeStr:
        if self._is_element_in_span_dict(self._span_dict_str):
            return self._span_dict_str
        subject_span = self._subject.span(self._subject.root)
        object_span = self._object.span(self._object_root)
        self._span_dict_str["S"] = self._subject.span_str(subject_span)
        self._span_dict_str["V"] = self._verb_root.text
        self._span_dict_str["O"] = self._object.span_str(object_span)
        return self._span_dict_str
    
    def _is_element_in_span_dict(self, dic:Union[ThirdSpanDictType,ThirdSpanDictTypeStr]) -> bool:
        if len(dic) == 0:
            return False
        return True
    
    @property
    def abbreviation(self) -> str:
        return "SVO"


class FourthSentencePattern(IBaseSentencePattern):
    def __init__(self, subject: Subject, verb: Verb, object: RootObject):
        self._subject = subject
        self._verb = verb
        self._object = object
        self._object_root = object.root
        self._subject_root = subject.root
        self._verb_root = verb.root
        self._span_dict:FourthSpanDictType = dict()
        self._span_dict_str:FourthSpanDictTypeStr = dict()

    @property
    def subject(self) -> "Subject": return self._subject
    @property
    def verb(self) -> "Verb": return self._verb
    @property
    def object(self) -> "RootObject": return self._object
    
    @property
    def spans(self) -> FourthSpanDictType:
        if self._is_element_in_span_dict(self._span_dict):
            return self._span_dict
        self._span_dict["S"] = self._subject.span(self._subject_root)
        self._span_dict["V"] = self._verb.root
        self._span_dict["O1"] = self._object.span(self._object_root)[0]
        self._span_dict["O2"] = self._object.span(self._object_root)[1]
        return self._span_dict
    
    @property
    def spans_str(self) -> FourthSpanDictTypeStr:
        if self._is_element_in_span_dict(self._span_dict_str):
            return self._span_dict_str
        subject_span = self._subject.span(self._subject.root)
        object_spans = self._object.span(self._object_root)
        self._span_dict_str["S"] = self._subject.span_str(subject_span)
        self._span_dict_str["V"] = self._verb.root.text
        self._span_dict_str["O1"] = self._object.span_str(object_spans)[0]
        self._span_dict_str["O2"] = self._object.span_str(object_spans)[1]
        return self._span_dict_str
    
    def _is_element_in_span_dict(self, dic:Union[FourthSpanDictType,FourthSpanDictTypeStr]) -> bool:
        if len(dic) == 0:
            return False
        return True
    
    @property
    def abbreviation(self) -> str:
        return "SVOO"


class FifthSentencePattern(IBaseSentencePattern):
    def __init__(self,
                 subject: Subject,
                 verb: Verb,
                 object: RootObject,
                 adjective: Adjective):

        self._subject = subject
        self._verb = verb
        self._object = object
        self._adjective = adjective
        self._object_root = object.root
        self._subject_root = subject.root
        self._verb_root = verb.root
        self._adjective_root = adjective.root
        self._span_dict = dict()
        self._span_dict_str = dict()

    @property
    def subject(self) -> "Subject": return self._subject
    @property
    def verb(self) -> "Verb": return self._verb
    @property
    def object(self) -> "RootObject": return self._object
    @property
    def adjective(self) -> "Adjective": return self._adjective
    
    @property
    def spans(self) -> FifthSpanDictType:
        if self._is_element_in_span_dict(self._span_dict):
            return self._span_dict
        self._span_dict["S"] = self._subject.span(self._subject_root)
        self._span_dict["V"] = self._verb.root
        self._span_dict["O"] = self._object.span(self._object_root)
        self._span_dict["C"] = self._adjective.span(self._adjective_root)
        return self._span_dict
    
    @property
    def spans_str(self) -> FifthSpanDictTypeStr:
        if self._is_element_in_span_dict(self._span_dict_str):
            return self._span_dict_str
        subject_span = self._subject.span(self._subject.root)
        object_spans = self._object.span(self._object_root)
        adjective_span = self._adjective.span(self._adjective_root)
        self._span_dict_str["S"] = self._subject.span_str(subject_span)
        self._span_dict_str["V"] = self._verb.root.text
        self._span_dict_str["O"] = self._object.span_str(object_spans)
        self._span_dict_str["C"] = self._adjective.span_str(adjective_span)
        return self._span_dict_str

    def _is_element_in_span_dict(self, dic:Union[FifthSpanDictType, FifthSpanDictTypeStr]) -> bool:
        if len(dic) == 0:
            return False
        return True
    
    @property
    def abbreviation(self) -> str:
        return "SVOC"

    

SentencePatternType: TypeAlias = FirstSentencePattern | SecondSentencePattern | ThirdSentencePattern | FourthSentencePattern | FifthSentencePattern

class SentencePatternDoc:
    def __init__(self, elements: RootElements):
        self.elements = elements

    @property
    def type(self) -> SentencePatternType: return self._classify_pattern_type()

    def _classify_pattern_type(self) -> SentencePatternType:
        """
        classify as one of the fifth sentence types

        Returns
        -------
        pattern: BaseSentencePatternInterface
        """

        subject = self.elements.subject
        verb = self.elements.verb
        adjective = self.elements.adjective
        rootobject = self.elements.rootobject
        
        if self.is_fifth_sentence(adjective=adjective, object=rootobject):
            return FifthSentencePattern(subject=subject, verb=verb, object=rootobject, adjective=adjective)

        if adjective.have_root:
            if rootobject.object_num == 1:
                pattern = FifthSentencePattern(subject=subject, verb=verb, object=rootobject, adjective=adjective)
            else:
                pattern = SecondSentencePattern(subject=subject, verb=verb, adjective=adjective)

        elif rootobject.object_num == 2:
            pattern = FourthSentencePattern(subject=subject, verb=verb, object=rootobject)
        elif rootobject.object_num == 1:
            pattern = ThirdSentencePattern(subject=subject, verb=verb, object=rootobject)
        else:
            pattern = FirstSentencePattern(subject=subject, verb=verb)
        return pattern

    def is_fifth_sentence(self, adjective: Adjective, object: RootObject) -> bool:
        if adjective.have_root == False:
            return False
        if adjective.root.pos == ADJ:
            for token in adjective.root.children:
                if token.dep == nsubj:
                    object.root.append(token)
                    return True
        return False


from abc import ABCMeta, abstractmethod
from typing import List,Dict
from spacy.tokens import Token
from sent_pattern.core.elements import Subject,Verb

class IBaseSentencePattern(metaclass=ABCMeta):
    """ All FifthSentencePattern Base Interface
    """
    @property
    @abstractmethod
    def subject(self) -> Subject:
        raise NotImplementedError()
    
    @property
    @abstractmethod
    def verb(self) -> Verb:
        raise NotImplementedError()
    
    @property
    @abstractmethod
    def spans(self) -> Dict[str, List[Token]]:
        raise NotImplementedError()
    
    @property
    @abstractmethod
    def spans_str(self) -> Dict[str, List[str]]:
        """
        span property str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def abbreviation(self) -> str:
        raise NotImplementedError()


from typing import Any, List, Optional
from spacy.tokens import Token
from sent_pattern.core.interface.Ielement import IRootElement


class Verb:
    def __init__(self, dep_list):
        self._dep_list = dep_list
        self._verb_root = self._get_root_verb()

    @property
    def root(self) -> Token:
        return self._verb_root

    @property
    def root_verb_children(self)-> List[Any]:
        child = [child for child in self._dep_list["ROOT"][0].children]
        return child

    def _get_root_verb(self) -> Optional[Token]:
        root = self._dep_list["ROOT"][0]
        if not root:
            return
        return root



import spacy
import sent_pattern
from typing import List, Tuple
import pytest

nlp = spacy.load("en_core_web_lg")
nlp.add_pipe("span_noun")
nlp.add_pipe("sent_pattern")


test_data = {
    "texts": [
        "The firm obtained that information without users' consent from a researcher who had been allowed by Facebook to deploy an app on the platform which harvested data from millions of its users.",
        "She makes me happy.",
        "I am hero",
        "I give you something",
        "The Eureka client handles all aspects of service instance registration and deregistration",
        "She goes to school"
        ],
    "pattern": ["SVO"],
    "spans": [
        {"SVO":["The firm", "obtained", "that information"]},
        {"SVOC":["She", "makes", "me", "happy"]},
        {"SVC": ["I", "am", "hero"]},
        {"SVOO": ["I", "give", "you", "something"]},
        {"SVO": ["The Eureka client", "handles", "all aspects of service instance registration and deregistration"]},
        {"SV": ["She", "goes", ]}
        ]
}

def make_test_data(texts: List[str], answers: List[str])-> List[Tuple[str,str]]:
    data = []
    for text, ans in zip(texts, answers):
        data.append((text, ans))
    return data

pattern_test_data = make_test_data(test_data["texts"], test_data["pattern"])
@pytest.mark.parametrize(('text', 'pattern'), pattern_test_data)
def test_pattern(text:str, pattern:str):
    doc = nlp(text)
    p = doc._.sentpattern
    assert p.abbreviation == pattern


pattern_test_data = make_test_data(test_data["texts"], test_data["spans"])
@pytest.mark.parametrize(('text', 'spans'), pattern_test_data)
def test_pattern_span(text:str, spans:str):
    doc = nlp(text)
    p = doc._.sentpattern
    if spans.get("SV") != None:
        data = spans.get("SV")
        s = data[0]
        v = data[1]
        assert s == p.spans_str["S"]
        assert v == p.spans_str["V"]
    elif spans.get("SVC") != None:
        data = spans.get("SVC")
        s = data[0]
        v = data[1]
        c = data[2]
        assert s == p.spans_str["S"]
        assert v == p.spans_str["V"]
        assert c == p.spans_str["C"]
    elif spans.get("SVO") != None:
        data = spans.get("SVO")
        s = data[0]
        v = data[1]
        o = data[2]
        assert s == p.spans_str["S"]
        assert v == p.spans_str["V"]
        assert o == p.spans_str["O"]
    elif spans.get("SVOO") != None:
        data = spans.get("SVOO")
        s = data[0]
        v = data[1]
        o1 = data[2]
        o = data[3]
        assert s == p.spans_str["S"]
        assert v == p.spans_str["V"]
        assert o1 == p.spans_str["O1"]
        assert o == p.spans_str["O2"]
    elif spans.get("SVOC") != None:
        data = spans.get("SVOC")
        s = data[0]
        v = data[1]
        o = data[2]
        c = data[3]
        print(p)
        assert s == p.spans_str["S"]
        assert v == p.spans_str["V"]
        assert o == p.spans_str["O"]
        assert c == p.spans_str["C"]


        
        


from typing import List, Union
from sent_pattern.core.elements.patterns import SentencePatternDoc
from sent_pattern.core.interface.Ipattern import IBaseSentencePattern
from sent_pattern.tags import create_dep_list, create_elements
from spacy.tokens import Doc
from spacy.language import Language


def get_sentpattern_type(doc: Doc)-> Union[IBaseSentencePattern,List[IBaseSentencePattern]]:
    pattern_docs = []
    if len(list(doc.sents)) > 1:
        for d in doc.sents:
            dep_list = create_dep_list(d)
            elements = create_elements(dep_list)
            pattern_doc = SentencePatternDoc(elements)
            pattern_docs.append(pattern_doc.type)
        return pattern_docs
    else:
        dep_list = create_dep_list(doc)
        elements = create_elements(dep_list)
        pattern_doc = SentencePatternDoc(elements)
    return pattern_doc.type

@Language.component("sent_pattern")
def create_sentence_pattern(doc: Doc):
    if not Doc.has_extension("sentpattern"):
        Doc.set_extension("sentpattern", getter=get_sentpattern_type)
    return doc



from .adjective import Adjective as Adjective
from .subject import Subject as Subject
from .verb import Verb as Verb
from .object import RootObject as RootObject
from .elements import RootElements as RootElements
from .elements import ElementOption as ElementOption
from .elements import CustomElements as CustomElements
from .elements import make_custom_elements as make_custom_elements
from .elements import make_root_elements as make_root_elements



from spacy.language import Language
from spacy.tokens import Doc


@Language.component("sub_text")
def create_subtexts(doc):
    # restore「,」index list
    punct_idxs = [0]
    subtexts = []
    for token in doc:
        if token.is_punct:
            punctype = token.morph.get("PunctType")
            punctype = "".join(punctype)
            if punctype == "Comm":
                punct_idxs.append(token.i + 1)
                subtexts.append(doc[punct_idxs[-2]:  punct_idxs[-1]])
    if subtexts:
        Doc.set_extension("sub_texts", default=subtexts, force=True)
    return doc



from typing import List, Optional
from spacy.tokens import Span, Token, SpanGroup, Doc

class PhraseData:
    PREP_LIST = [
        "about", "after", "as", "at", "before", "by", "for", "from", "in", "into", "near", "of", "on", "over", "to",
        "under", "with", "above", "across", "against", "along", "among", "around", "behind", "below", "beside",
        "between", "despite", " during", "except", " like", "off", "onto", "since", "than", "throughout", "toward",
        "toward", "until", "via", "within", "without", "worth", "anti", "beneath", " beyond", " concering", "down",
        "excluding", "including", "inside", "outside", "past", "per", "pro", "regarding", "underneath", "unlike", "up",
        "versus", "aboard", "alongside", "amid", "but", "circa", "failing", "fllowing", "minus", "opposite", "out", "plus",
        "times", "abaft", "absent", "apropos", "astride", "athwart", "given", "next", "pace", "qua", "save",
    ]


class PrepPhrase():

    def __init__(self, doc:Doc):
        def is_prep(token): return token.text in PhraseData.PREP_LIST
        Token.set_extension("is_prep", getter=is_prep, force=True)
        self._doc = doc
        self.__prep_phrase = self._get_prep_phrase()
    
    @property
    def prep_phrase(self) -> Optional[List[Span]]:
        return self.__prep_phrase
    
    def _get_prep_phrase(self) -> Optional[List[Span]]:
        """method to extract a noun including prep 

        Args:
            doc (Doc): spacy.tokens.Doc class

        Returns:
            Optional[List[Span]]: return a noun including prep 
        """        
        prep_noun = "prep_noun"
        spans = []
        for token in self._doc:
            subs = [sub for sub in token.subtree]
            if token.dep_ == "prep":
                start = subs[0].i
                end = subs[-1].i+1
                span = Span(self._doc, start, end, label=prep_noun)
                spans.append(span)
        return spans


from spacy.language import Language
from spacy.tokens import Doc


@Language.component("span_noun")
def create_span_noun(doc: Doc):
    """
    process noun_chunks as a single noun
    """
    noun_list = list(doc.noun_chunks)
    with doc.retokenize() as retokenizer:
        for span in noun_list:
            retokenizer.merge(span)
    return doc



from typing import TypeAlias, Dict, List, Optional,Union
from spacy.tokens import Token

DepLemmaListType: TypeAlias = Dict[str, List[Optional[Token]]]
AdjectiveSpanType: TypeAlias = Optional[List[Token]]
AdjectiveSpanStrType: TypeAlias = Optional[str]
AdjectiveRootType: TypeAlias = Union[Token, str]
SubjectSpanType: TypeAlias = List[Token]
SubjectRootType: TypeAlias = Token
ObjectRootType: TypeAlias = List[Optional[Token]]
ObjectSpanType: TypeAlias = Optional[List[Union[List[Token], Token]]]


FirstSpanDictType: TypeAlias = Dict[str, Union[Token, List[Token]]]
FirstSpanDictTypeStr: TypeAlias = Dict[str, Optional[str]]


SecondSpanDictType: TypeAlias = Dict[str, List[Optional[Union[Token, str, List[Token]]]]]
SecondSpanDictTypeStr: TypeAlias = Dict[str, Optional[str]]

ThirdSpanDictType: TypeAlias = Dict[str, List[Optional[Union[Token, str, List[Token]]]]]
ThirdSpanDictTypeStr: TypeAlias = Dict[str, Optional[str]]

FourthSpanDictType: TypeAlias = Dict[str, Optional[List[Union[List[Token], Token]]]]
FourthSpanDictTypeStr: TypeAlias = Dict[str, Optional[str]]

FifthSpanDictType: TypeAlias = Dict[str, Optional[List[Union[List[Token], Token]]]]
FifthSpanDictTypeStr: TypeAlias = Dict[str, Optional[str]]


import spacy
from sent_pattern import tags

nlp = spacy.load("en_core_web_lg")
nlp.add_pipe("span_noun")
nlp.add_pipe("sent_pattern")


def test_pipe_added_to_nlp():
    assert nlp.pipe_names[-1] == "sent_pattern"



from typing import List,  Optional, Union
from sent_pattern.core.interface.Ielement import IRootElement
from sent_pattern.core.type import DepLemmaListType
from sent_pattern.core.type import ObjectSpanType, ObjectRootType,  DepLemmaListType


class RootObject(IRootElement):
    DEP = [
        "pobj",
        "obj",
        "dative",
        "dobj"
    ]

    def __init__(self, dep_list: DepLemmaListType):
        self._dep_list = dep_list
        self._object_root = self._get_root()

    @property
    def dep_list(self) -> DepLemmaListType:
        return self._dep_list

    @property
    def root(self) -> ObjectRootType:
        return self._object_root
    
    @property
    def object_num(self) -> int:
        return len(self.root)
    
    def span(self,root:ObjectRootType) ->ObjectSpanType: 
        return self._get_span(root)
    
    def span_str(self, spans: ObjectSpanType) -> Optional[Union[List[str],str]]:
        """If the target word is included in the sentence, the tokens are concatenated into a string

        Args:
            spans (ObjectSpanType): tokens related to the target word token

        Returns:
            Optional[Union[List[str],str]]: argument spans converted to string
        """        
        if len(spans) == 0:
            return
        elif len(spans) == 1:
            return " ".join([token.text for token in spans[0]])
        elif len(spans) == 2:
            return [" ".join([token.text for token in spans[0]])," ".join([token.text for token in spans[1]])]
        return

    def _get_root(self) -> ObjectRootType:
        """If the child token's dep of the verb in the sentence is included in the DEP list, return a list of the tokens

        Returns:
            ObjectRootType: List of tokens matched to DEP list
        """        
        objects = []
        root_verb = self._dep_list["ROOT"][0]
        child = [child.dep_ for child in root_verb.children]
        for object_dep in RootObject.DEP:
            if object_dep in child:
                objects.append(self._dep_list[object_dep][0])
        return objects

    def _get_span(self, root:ObjectRootType) ->ObjectSpanType:
        """extract tokens related to the target word token

        Args:
            root (ObjectRootType): Token corresponding to sentence type "O" (ex. ThirdSentencePattern(SVO))

        Returns:
            ObjectSpanType: tokens related to the target word token
        """        
        if len(root) == 0:
            return
        if len(root) == 1:
            return [[token for token in root[0].subtree]]
        elif len(root) == 2:
            return [[token for token in root[0].subtree], [token for token in root[1].subtree]]
        return
    
    
        


import spacy
import sent_pattern
from typing import List, Tuple
import pytest

nlp = spacy.load("en_core_web_lg")
nlp.add_pipe("span_noun")
nlp.add_pipe("sent_pattern")

test_data = {
    "texts": [
        "The firm obtained that information without users' consent from a researcher who had been allowed by Facebook to deploy an app on the platform which harvested data from millions of its users. She makes me happy.",
        "I like you"
        ],
    "pattern": ["SVO SVOC", "SVO"],
}

def make_test_data(texts: List[str], answers: List[str])-> List[Tuple[str,str]]:
    data = []
    for text, ans in zip(texts, answers):
        data.append((text, ans))
    return data

pattern_test_data = make_test_data(test_data["texts"], test_data["pattern"])
@pytest.mark.parametrize(('text', 'pattern'), pattern_test_data)
def test_pattern(text:str, pattern:str):
    doc = nlp(text)
    ps = pattern.split()
    patterns = doc._.sentpattern
    if len(list(doc.sents)) == 1:
        assert patterns.abbreviation == ps[0]
        return
    for s, p in zip(ps, patterns):
        assert p.abbreviation == s


from collections import defaultdict
from spacy.tokens import Doc
from sent_pattern.core.elements.elements import RootElements, make_root_elements
from sent_pattern.core.elements.patterns import SentencePatternDoc
from sent_pattern.core.type import DepLemmaListType


def create_dep_list(doc: Doc) -> DepLemmaListType:
    """
    Put into dict for each dep of Token
    Parameters
    ----------
    doc: spacy.tokens.Doc
        doc after nlp initialization

    Returns
    -------
    dep_list: Dict[str, Optional[Token]]
    """
    dep_list = defaultdict(list)
    for token in doc:
        dep_list[token.dep_].append(token)
    return dict(dep_list)


def create_elements(dep_list: DepLemmaListType) -> RootElements:
    """
    Parameters
    ----------
    dep_list : DepLemmaListType
        Put into dict for each dep of Token

    Returns
    -------
    elements: RootElements
        classes have subject, verb, adjective, and object
    """
    elements = make_root_elements(
        dep_list=dep_list)
    return elements


def create_sent_pattern(elements: RootElements) -> SentencePatternDoc:
    """
    Parameters
    ----------
    elements : RootElements
        classes have subject, verb, adjective, and object

    Returns
    -------
    pattern : SentencePatternFactory
    """
    pattern = SentencePatternDoc(elements)
    return pattern



from sent_pattern.core.interface.Ielement import IRootElement
from sent_pattern.core.type import AdjectiveRootType, AdjectiveSpanStrType, AdjectiveSpanType, DepLemmaListType


class Adjective(IRootElement):
    DEP = [
        "acomp",
        "ccomp",
        "complm",
        "pcomp",
        "xcomp",
        "attr",
        "oprd",
        "npadvmod"
    ]

    def __init__(self, dep_list:DepLemmaListType):
        self._dep_list = dep_list
        self._adjective_root = self._get_root()
    
    @property
    def root(self) -> AdjectiveRootType:
        """
        Return Adjective (C)

        Returns:
            - Adjective(Token) | ""(str)
        """
        return self._adjective_root

    def span(self,root:AdjectiveRootType) -> AdjectiveSpanType: return self._get_span(root)
    
    def span_str(self, spans:AdjectiveSpanType) -> AdjectiveSpanStrType:
        if len(spans) == 0:
            return
        else:
            return " ".join([token.text for token in spans if token.dep_ != "nsubj"])
    
    @property
    def have_root(self) -> bool:
        """Whether or not have an adjective
        Params:

        Returns:
            - True: has adjective
            - False: not adjective
        """
        return bool(self.root)

    def _get_root(self) -> AdjectiveRootType:
        """
        get root adjective
        Return
                adjctive(Token) or empty string(string)
        """
        root_verb = self._dep_list["ROOT"][0]
        child = [child.dep_ for child in root_verb.children]

        for speech in self.DEP:
            if speech in child:
                adjective = self._dep_list[speech]
                return adjective[0]
        return ""

    def _get_span(self, root:AdjectiveRootType)->AdjectiveSpanType:
        """method that returns a list of subtrees of the root of the argument

        Args:
            root (AdjectiveRootType): _description_

        Returns:
            AdjectiveSpanType:  Type[List[Token]] | Type[None]
        """        
        if type(root) == str:
            return
        adje_list = [token for token in root.subtree]
        return adje_list

    


from typing import List, Optional, Union
from sent_pattern.core.elements import Subject, Verb, Adjective, RootObject
from spacy.tokens import Doc
from sent_pattern.core.elements.sub.phrase import PrepPhrase
from sent_pattern.core.elements.sub.relative import RelativeClause
from sent_pattern.core.type import DepLemmaListType

def make_root_elements(dep_list: DepLemmaListType) -> "RootElements":
    """
    create instance of self
    Parameters
    ----------
    dep_list : type.DepLemmaListType

    Returns
    -------
    RootElements : self
    """
    subject = Subject(dep_list)
    verb = Verb(dep_list)
    adjective = Adjective(dep_list)
    rootobject = RootObject(dep_list)

    return RootElements(subject, verb, adjective, rootobject)


def make_custom_elements(doc:Doc, dep_list: DepLemmaListType,  opt:str = "all") -> "CustomElements":
    """_summary_

    Args:
        dep_list (DepLemmaListType): dict for each dep of Token
        doc (Doc): spacy.tokens.Doc
        option (str): _description_

    Returns:
        CustomElements: _description_
    """        
    subject = Subject(dep_list)
    verb = Verb(dep_list)
    adjective = Adjective(dep_list)
    rootobject = RootObject(dep_list)
    options = ElementOption(option=opt)
    custom_elements = CustomElements(doc, subject, verb, adjective, rootobject, option=options)
    return custom_elements


class RootElements:
    """
    this class have just subject, verb, adjective and object
    """
    def __init__(self, 
            subject: Subject, 
            verb: Verb, 
            adjective: Adjective, 
            rootobject: RootObject):

        self.subject = subject
        self.verb = verb
        self.adjective = adjective
        self.rootobject = rootobject


class ElementOption:
    _all = "all"
    Prep = "prep"
    Relcl = "relcl"
    
    def __init__(self, option: Union[List[str],str]) -> None:
        self._option = option

    @classmethod
    def has_option(cls, option: str) -> bool:
        """returns a bool indicating whether the option is configurable or not

        Args:
            option (str): option str (etc. prep, relcl)

        Returns:
            bool: whether the option is configurable or not
        """
        opt_list = cls.option_list()
        return option in opt_list

    @classmethod
    def option_list(cls) -> List[str]:
        """returns a list of what can be set as options

        Returns:
            str: list of options
        """
        return [cls.Prep, cls.Relcl, cls._all]

    def __str__(self) -> str:
        if type(self._option) is str:
            return self._option
        return " ".join([e for e in self._option])

class CustomElements(RootElements):
    def __init__(self,
            doc: Doc,
            subject: Subject, 
            verb: Verb, 
            adjective: Adjective, 
            rootobject: RootObject, 
            option: ElementOption):

        super().__init__(subject, verb, adjective, rootobject)
        self._option = self._is_valid_option(option.__str__())
        self._doc = doc
        self.__relcl = self._get_relcl()
        self.__prep = self._get_prep()

    def _is_valid_option(self, option: str) -> str:
        """determine if the argument option is a configurable option

        Args:
            option (str): _description_

        Raises:
            ValueError: _description_

        Returns:
            str: _description_
        """        
        try:
            if ElementOption.has_option(option):
                return option
            else:
                raise ValueError
        except ValueError:
            raise ValueError(f"invalid option. your option is {option} valid option: [{ElementOption.option_list()}]")
    
    @property
    def all(self):
        return [self.relcl, self.prep]
    
    @property
    def relcl(self) -> Optional["RelativeClause"]:
        if self._option == "all" or self._option == ElementOption.Relcl:
            return self.__relcl
        return
    
    @property
    def prep(self) ->Optional["PrepPhrase"]:
        return self.__prep
    
    def _get_relcl(self) -> Optional[RelativeClause]:
        relcl = RelativeClause(self._doc)
        if relcl.relcl_section:
            return relcl
        return 
    
    def _get_prep(self) -> Optional["PrepPhrase"]:
        prep = PrepPhrase(self._doc)
        if prep.prep_phrase:
            return prep
        return


from typing import List, Optional
from spacy.tokens import Doc, Span


class RelativeClause():
    def __init__(self, doc: Doc) -> None:
        self._doc = doc
        self.relcl_section = self._get_relative(doc)

    def _get_relative(self, doc: Doc) -> Optional[List[Span]]:
        """method to extract relational clauses from Doc class and return them as a list

        Args:
            doc (Doc): spacy.tokens.Doc class

        Returns:
            Optional[List[Span]]: return spacy.tokens.Span list
        """        
        spans = []
        for token in doc:
            if token.dep_ == "relcl":
                subtree = [t for t in token.subtree]
                start = doc[subtree[0].i].i
                end = doc[subtree[-1].i].i+1
                span = Span(doc, start, end, label="relcl")
                spans.append(span)
        return spans



#!/usr/bin/python3
import setuptools
import sent_pattern
with open("README.md", "r") as fh:
    long_description = fh.read()


def get_lines(path):
    with open(path) as f:
        return f.readlines()


factory = [
    "sent_pattern = sent_pattern.pipelines.sent_pattern:create_sentence_pattern",
    "span_noun = sent_pattern.pipelines.span_noun:create_span_noun"
]

setuptools.setup(
    name='sent-pattern',
    version=sent_pattern.__version__,
    author='Shunpei Nakayama',
    license="MIT",
    author_email='hoku804049@gmail.com',
    description='project that interpret English sentences to improve your ability to read English sentences correctly.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/lll-lll-lll-lll/sent-pattern',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.8',
    entry_points={
        "spacy_factories": factory,
    },
    install_requires=get_lines("requirements.txt"),
)



from sent_pattern.core.interface.Ielement import IRootElement
from sent_pattern.core.type import DepLemmaListType
from typing import Optional
from sent_pattern.core.type import SubjectRootType, SubjectSpanType,  DepLemmaListType


class Subject(IRootElement):
    DEP = [
        "nsubj",
        "nsubjpass",
        "csubj",
        "csubjpass"
    ]

    def __init__(self, dep_list: DepLemmaListType):
        self._dep_list = dep_list
        self._subject_root = self._get_root()

    @property
    def root(self) -> SubjectRootType:
        """
        return subject
        Params:

        Returns:
            - _get_root(self): (SubjectRootType)
        """
        return self._subject_root

    def span(self,root:SubjectRootType) -> SubjectSpanType:
        """
        Span summarizing subtrees of elements
        """
        return self._get_span(root)
    
    def span_str(self, spans:SubjectSpanType) -> str:
        """
        span property str
        """
        return " ".join([token.text for token in spans])

    def _get_root(self) -> SubjectRootType:
        subj_type = self._get_subj_type()
        root_subject = self._dep_list[subj_type][0]
        return root_subject

    def _get_subj_type(self) -> Optional[str]:
        root_verb = self._dep_list["ROOT"][0]
        verb_child_dep = [
            child.dep_ for child in root_verb.children]
        for subj in Subject.DEP:
            if subj in verb_child_dep:
                return subj
    
    def _get_span(self, root:SubjectRootType) -> SubjectSpanType:
        return [token for token in root.subtree]
    
    



from . import span_noun
from .import sent_pattern
from . import sub_texts
from .span_noun import create_span_noun as create_span_noun
from .sub_texts import create_subtexts as create_subtexts
from .sent_pattern import create_sentence_pattern as create_sentence_pattern



from abc import abstractmethod, ABCMeta
from spacy.tokens import Token
from typing import Union

class IRootElement(metaclass=ABCMeta):
    @property
    @abstractmethod
    def root(self) -> Union[Token, str]:
        raise NotImplementedError()

    @abstractmethod
    def span(self):
        raise NotImplementedError()

    @abstractmethod
    def span_str(self):
        raise NotImplementedError()


from sent_pattern.core.elements.elements import make_custom_elements
from sent_pattern.core.elements.sub.phrase import PrepPhrase
import spacy
from sent_pattern import tags


nlp = spacy.load("en_core_web_lg")

def test_custom_phrase():
    text = "The Eureka client handles all aspects of service instance registration and deregistration"
    doc = nlp(text)
    dep_list = tags.create_dep_list(doc)
    custom = make_custom_elements(doc,dep_list ,opt="prep")
    assert custom.prep.__class__ == PrepPhrase
    

def test_custom_relcl(): 
    text = "The Eureka client handles all aspects of service instance registration and deregistration"
    doc = nlp(text)
    dep_list = tags.create_dep_list(doc)
    custom = make_custom_elements(doc,dep_list ,opt="relcl")
    print(custom.relcl)
    assert custom.relcl == None
